#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import abc
import glob
import pathlib
import re
from typing import Dict, Tuple, Union

import chevron

from .deliverer import Destination
from .destinations.interfaces import ProductMetadataIF, SpooledProduct


class ProductMetadata(ProductMetadataIF):
    """
    Represents the metadata about a given product.
    """

    @property
    def project(self) -> str:
        return self.project_code

    telescope: str
    project_code: str
    type: str
    segment: str
    mous_id: str
    date: str

    # These templates are taken from
    # https://open-confluence.nrao.edu/pages/viewpage.action?spaceKey=SPR&title=Delivery+Directory+Improvements
    DIRECTORY_TEMPLATES = {
        "NRAO": "$DELIVERY_ROOT/$PROJECT_CODE/$PIPELINE_SPEC/$PRODUCT",
        "ALMA": "$DELIVERY_ROOT/$PROJECT_CODE/$MOUS_ID/$PIPELINE_SPEC/$PRODUCT",
        "VLBA": "$DELIVERY_ROOT/$PROJECT_CODE/$SEGMENT/$PRODUCT",
        "GMVA": "$DELIVERY_ROOT/$PROJECT_CODE/$SEGMENT/$PRODUCT",
        "GBT":  "$DELIVERY_ROOT/$PROJECT_CODE/$PROJECT_CODE_$SEGMENT/raw/$PRODUCT",
    }

    def __init__(self, telescope: str, project_code: str, type: str, segment: str, mous_id: str, date: str):
        self.telescope = telescope
        self.project_code = project_code
        self.type = type
        self.segment = segment if segment != project_code else None
        self.mous_id = mous_id
        self.date = date if date else "unknown_date"

    @classmethod
    def from_json(cls, json: dict) -> "ProductMetadata":
        return cls(
            json["telescope"], json["project_code"], json["type"], json["segment"], json["mous_id"], json["date"]
        )

    def to_path(self) -> str:
        """
        Convert this product to a path, following the instructions at
        https://open-confluence.nrao.edu/pages/viewpage.action?spaceKey=SPR&title=Delivery+Directory+Improvements

        :return: a path string for this product
        """
        return chevron.render(self.template_to_mustache(self.choose_template()), self.__json__())

    def choose_template(self) -> str:
        """
        Find the right template for this instrument.

        If we don't have a special template for this instrument, we'll use the NRAO template.
        :return:
        """
        # this circumlocution gets us what you asked for, or NRAO, but never nothing
        return ProductMetadata.DIRECTORY_TEMPLATES.get(self.telescope, ProductMetadata.DIRECTORY_TEMPLATES["NRAO"])

    def template_to_mustache(self, template: str):
        # first, remove the prefix, as we don't need it
        template = template.replace("$DELIVERY_ROOT/", "")

        # second, replace $FOO with {foo}
        return re.sub(r"\$([A-Z_]+[A-Z])", lambda match: "{{" + match.group(1).lower() + "}}", template)

    @property
    def pipeline_spec(self) -> str:
        """
        The pipeline specification for this product. May be overridden by a subclass.

        For execution blocks, $PIPELINE_SPEC is observation.$DATE
        For calibrations, $PIPELINE_SPEC is calibration_pipeline.$DATE
        For images, $PIPELINE_SPEC is imaging_pipeline.$DATE
        :return:
        """
        if self.type == "execblock":
            return f"observation.{self.date}"
        elif self.type == "calibration":
            return f"calibration_pipeline.{self.date}"
        elif self.type == "image":
            return f"imaging_pipeline.{self.date}"
        else:
            return f"pipeline.{self.date}"

    def __json__(self) -> dict:
        return dict(
            telescope=self.telescope,
            project_code=self.project_code,
            type=self.type,
            segment=self.segment,
            mous_id=self.mous_id,
            date=self.date,
            pipeline_spec=self.pipeline_spec,
        )

    def __str__(self):
        return f"{self.type} {self.pipeline_spec}"


class ArchiveProduct(SpooledProduct):
    """
    An archive product is either a science product or an ancillary product from the archive.
    Because it originates in the archive, there is certain metadata about it which we are in
    possession of. This metadata determines how the delivery path is constructed.

    For more information, see ProductMetadata.
    """

    def __init__(self, path: pathlib.Path, metadata: Union[dict, ProductMetadata]):
        super().__init__(path)
        self._metadata = metadata if type(metadata) == ProductMetadata else ProductMetadata.from_json(metadata)

    def deliver_to(self, destination: Destination):
        # if we add ourself directly, we'll get an extra directory with a wonky name
        # looping this way instead ensures we don't get extras
        #
        # however, there is a catch: we have to exclude the directory name if there is only one item in it.
        # but only at the top level, which is why we are doing it here and not in add_directory.
        for item in self.path.iterdir():
            if self.directory_contains_single_item(item):
                self.product_destination(destination).add_directory(item, ".")
            else:
                self.product_destination(destination).add_directory(item, item.name)

    def product_destination(self, destination: Destination) -> Destination:
        return destination / self.metadata.to_path()

    @staticmethod
    def directory_contains_single_item(path: pathlib.Path) -> bool:
        """
        Returns true if there is just a single item under a certain path.

        :param path:  the path to check
        :return: true if the path has just one item
        """
        # The idea here is that we're going to call next() on the path iterator to ascertain this.
        # If we can call next() successfully just once, we have our answer. So we need to try to call
        # it twice, and return false if the first one fails or the second one succeeds.

        # We're doing this instead of making a list of items, because sometimes archive retrievals are
        # enormous and we don't want to pay for generating the entire list.
        iterator = path.iterdir()

        # so first, we want to see if there is at least one item in the directory
        try:
            next(iterator)
        except StopIteration:
            # here we are returning false because the iterator actually had zero path items in it
            return False

        # if we got here, the first call succeeded, and now we're going to return true if the second one fails.
        try:
            next(iterator)

            # it didn't fail, so there are at least two items in this path
            return False
        except StopIteration:
            # it did fail, so there is in fact only one item in this path
            return True

    def __str__(self):
        return str(self.metadata)

    @property
    def metadata(self) -> ProductMetadataIF:
        return self._metadata


class Calibration(SpooledProduct):
    """
    This class is intended for use when we do not have product metadata to use.
    Probably will need to be refactored.
    """

    def __init__(self, path: pathlib.Path):
        super().__init__(path)
        self._metadata = self.parse_metadata()

    def deliver_to(self, destination: Destination):
        """
        List of all files typically found in a calibration spool dir.
        """
        # X - not relevant file (exclude from tar

        # products/
        # |-- 21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt
        # |-- 21A-409.sb39530507.eb39564922.59311.953874212966.ms.flagversions.tgz
        # |-- casa_commands.log
        # |-- casa_piperestorescript.py
        # |-- casa_pipescript.py
        # |-- flux.csv
        # |-- [X] oussid.J0319+4130_bp.A_band.cont.I.alpha.error.fits
        # |-- [X] oussid.J0319+4130_bp.A_band.cont.I.alpha.fits
        # |-- [X] oussid.J0319+4130_bp.A_band.cont.I.mask.fits
        # |-- [X] oussid.J0319+4130_bp.A_band.cont.I.pb.tt0.fits
        # |-- [X] oussid.J0319+4130_bp.A_band.cont.I.tt0.fits
        # |-- [X] oussid.J0319+4130_bp.A_band.cont.I.tt1.fits
        # |-- [X] oussid.J0552+0313_ph.A_band.cont.I.alpha.error.fits
        # |-- [X] oussid.J0552+0313_ph.A_band.cont.I.alpha.fits
        # |-- [X] oussid.J0552+0313_ph.A_band.cont.I.mask.fits
        # |-- [X] oussid.J0552+0313_ph.A_band.cont.I.pb.tt0.fits
        # |-- [X] oussid.J0552+0313_ph.A_band.cont.I.tt0.fits
        # |-- [X] oussid.J0552+0313_ph.A_band.cont.I.tt1.fits
        # |-- [X] pipeline-20210409T164741/
        # |-- pipeline_aquareport.xml
        # |-- PPR.xml
        # |-- [X] reprocess_flag_file.fits
        # |-- unknown.hifv_cal.auxproducts.tgz
        # |-- unknown.hifv_cal.pipeline_manifest.xml
        # |-- unknown.hifv_cal.pprequest.xml
        # |-- unknown.session_1.caltables.tgz
        # `-- [X] weblog.tgz
        # let's parse the metadata and make some mistakes
        project_code, pipeline_id = self.parse_metadata()

        # change to the pipeline directory
        pipedir = destination / project_code / f"calibration_pipeline.{pipeline_id}"

        # let's omit some junk
        with pipedir.excluding(".fits", ".tgz") as nofits:
            # let's deliver everything that isn't a fits or tgz file
            for file in self.path.iterdir():
                nofits.add_path_entry(file, file.name)

        # add the weblog manually because it was masked above
        weblog = self.path / "weblog.tgz"
        if weblog.exists():
            pipedir.add_file(weblog, "weblog.tgz")

    def parse_metadata(self) -> Tuple[str, str]:
        """
        Scan through the calibration products and collect information about the calibration

        :return: Dict of calibration metadata

        TODO: Collect metadata from calapply file name: project code, CASA pipeline run identification numbers
        """
        # There should only be a single `.ms.calapply.txt` file
        calapply_file, *_ = list(self.path.glob("**/*.ms.calapply.txt"))

        # The calapply file has a name like 21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt
        # For our purposes, we need the project code (21A-409) and the pipeline identifier (59311.953874212966).
        #
        # This regex obtains the project code (before the first dot), and then we look for two dot-separated groups
        # before the expected ending (ms.calapply.txt). Then we glue the second and third items back together to
        # recreate the pipeline identifier.
        parts = re.search(
            r"^(?P<project_code>[^.]+).*\.(?P<mjd>[^.]+)\.(?P<time>[^.]+)\.ms\.calapply\.txt$",
            calapply_file.name,
        )
        return parts["project_code"], f"{parts['mjd']}.{parts['time']}"

    @property
    def metadata(self) -> ProductMetadataIF:
        return self._metadata

    def __str__(self):
        # TODO: Add a uniquely defining piece of metadata to the str
        return f"calibration"
